# -*- coding: utf-8 -*-
import heimdall
from ..decorators import get_database, create_database
from .json import is_url, _create_tree, _tree2object
from urllib.parse import urlparse
from urllib.request import urlopen
try:
    from yaml import safe_load, dump
    _installed = True
except ModuleNotFoundError:  # pragma: no cover
    _installed = False

"""
Provides connectors to HERA-formatted YAML files.

This module defines input and output connectors to databases composed in full or in part of YAML files following the HERA schema.

* :py:class:`heimdall.connectors.yaml.getDatabase` is the input connector
* :py:class:`heimdall.connectors.yaml.serialize` is the output connector

:copyright: The pyHeimdall contributors.
:licence: Afero GPL, see LICENSE for more details.
:SPDX-License-Identifier: AGPL-3.0-or-later
"""  # nopep8: E501


def check_available():
    if not _installed:
        raise ModuleNotFoundError("Module 'pyyaml' required.")
    return _installed


@get_database('hera:yaml')
def getDatabase(**options):
    r"""Imports a database from a HERA YAML file

    :param \**options: Keyword arguments, see below.
    :Keyword arguments:
        * **url** (:py:class:`str`) -- Local ou remote path of the YAML file to import
        * **format** (:py:class:`str`, optional) Always ``hera:yaml``
        * **encoding** (:py:class:`str`, optional, default: ``utf-8``) -- ``url`` file encoding
    :return: HERA element tree
    :rtype: :py:class:`xml.etree.ElementTree.Element`

    Usage example: ::

      >>> import heimdall
      >>> tree = heimdall.getDatabase(format='hera:yaml', url='some/input.yaml')
      >>> # ... do stuff ...

    .. CAUTION::
       For future compability, this function shouldn't be directly called; as shown in the usage example above, it should only be used through :py:class:`heimdall.getDatabase`.
    """  # nopep8: E501
    check_available()  # breaks if not
    url = options['url']
    encoding = options.get('encoding', 'utf-8')
    if not is_url(url):
        with open(url, 'r') as f:
            data = safe_load(f)
    else:
        with urlopen(url) as response:
            data = safe_load(response.read().decode(encoding))
    return _create_tree(data or dict())


@create_database('hera:yaml')
def serialize(tree, url, **options):
    r"""Serializes a HERA elements tree into a YAML file

    :param tree: HERA elements tree
    :param url: Path of the YAML output file
    :param \**options: Keyword arguments, see below.

    :Keyword arguments:
        * **style** (``str``, default: ``block``) -- YAML flow style for
          output file. Valid values are ``flow`` of ``block``.
          Any value other than ``flow`` will be interpreted as ``block``.
        * **sort** (``bool``, default: ``False``) -- If ``True``, keys will be
          sorted in output file; if ``False`` keys will be written to ``url``
          as they are found in ``tree``.
    """
    check_available()  # breaks if not
    data = _tree2object(tree)
    # write data to file
    style = options.get('style', 'block')
    style = (style == 'flow')
    sort = options.get('sort', False)
    with open(url, 'w', encoding='utf-8') as f:
        dump(data, f, default_flow_style=style, sort_keys=sort)



# -*- coding: utf-8 -*-
from xml.etree.ElementTree import Element, QName
from heimdall.mixins import *


"""
Provides custom element classes.

:copyright: The pyHeimdall contributors.
:licence: Afero GPL, see LICENSE for more details.
:SPDX-License-Identifier: AGPL-3.0-or-later
"""


class Root(Node, Element):
    """HERA element tree root
    """
    def __init__(self, **kwargs):
        xml_schema = 'http://www.w3.org/2001/XMLSchema-instance'
        xsd = 'https://gitlab.huma-num.fr/datasphere/hera/schema/schema.xsd'
        qname = QName(xml_schema, 'schemaLocation')
        kwargs[qname] = xsd  # NOTE: maybe we could NOT force this?
        super().__init__('hera', **kwargs)

    def get_container(self, tag):
        for c in self.children:
            if c.tag == tag:
                return c
        return None


class Item(Node, Element):
    """Item custom element
    """
    def __init__(self, **kwargs):
        super().__init__('item', **kwargs)

    @property
    def metadata(self):
        return [e for e in self.iter() if e is not self]


class Metadata(Node, Element):
    """Metadata custom element
    """
    def __init__(self, **kwargs):
        super().__init__('metadata', **kwargs)


class Property(Documentable, Typable, Node, Element):
    """Property custom element
    """
    def __init__(self, **kwargs):
        super().__init__('property', **kwargs)


class Entity(Documentable, Node, Element):
    """Entity custom element
    """
    def __init__(self, **kwargs):
        super().__init__('entity', **kwargs)

    @property
    def attributes(self):
        return [e for e in self.children if e.tag == 'attribute']

    def append(self, child):
        super().append(child)
        if type(child) is Attribute:
            child.entity = self


class Attribute(Documentable, Typable, Repeatable, Node, Element):
    """Attribute custom element
    """
    def __init__(self, **kwargs):
        super().__init__('attribute', **kwargs)


class Builder(object):

    TAGS = {
        'hera': Root,
        'item': Item,
        'metadata': Metadata,
        'property': Property,
        'entity': Entity,
        'attribute': Attribute,
        }
    root = None
    parent = None

    stack = list()

    @property
    def current(self):
        try:
            return self.stack[-1]
        except IndexError:
            return None

    def start(self, tag, attrib):
        """Called for each opening ``tag`` and their attributes ``attrib``
        """
        try:
            element = self.TAGS[tag](**attrib)
        except KeyError:
            element = Element(tag, **attrib)
        if self.parent is None:
            self.root = element
        else:
            self.parent.append(element)
        self.parent = element
        self.stack.append(element)

    def end(self, tag):
        """Called for each closing ``tag``
        """
        element = self.stack.pop()
        self.parent = self.current

    def data(self, data):
        """Called for each element ``data`` (ie. text)
        """
        self.current.text = data

    def close(self):
        """Called when all data has been parsed
        """
        return self.root


__copyright__ = "Copyright the pyHeimdall contributors."
__license__ = 'AGPL-3.0-or-later'
__all__ = [
    'Builder',
    'Root',
    'Item', 'Metadata',
    'Property',
    'Entity', 'Attribute',

    '__copyright__', '__license__',
    ]



# -*- coding: utf-8 -*-
from heimdall.util import (
    get_language, get_nodes, get_node,
    create_nodes, create_node,
    )

"""
Provides mixins for custom element classes.

:copyright: The pyHeimdall contributors.
:licence: Afero GPL, see LICENSE for more details.
:SPDX-License-Identifier: AGPL-3.0-or-later
"""


class Documentable(object):

    @property
    def name(self):
        return self._icanhaz_get_dict('name')

    @name.setter
    def name(self, value):
        self._icanhaz_set_list('name', value)

    @name.deleter
    def name(self):
        self._icanhaz_del('name')

    @property
    def description(self):
        return self._icanhaz_get_dict('description')

    @description.setter
    def description(self, value):
        self._icanhaz_set_list('description', value)

    @description.deleter
    def description(self):
        self._icanhaz_del('description')

    @property
    def uri(self):
        return self._icanhaz_get_list('uri')

    @uri.setter
    def uri(self, values):
        if ((values is not None)
           and (type(values) is not list)
           and (type(values) is not str)):
            bad = str(type(values))
            raise TypeError(f"Argument must be list or str, got '{bad}'")
        self._icanhaz_set_list('uri', values)

    @uri.deleter
    def uri(self):
        self._icanhaz_del('uri')

    def _icanhaz_get_dict(self, tag):
        values = dict()
        for node in get_nodes(self, tag):
            lang = get_language(node)
            alternatives = values.get(lang, list())
            alternatives.append(node.text)
            values[lang] = alternatives
        return values

    def _icanhaz_get_list(self, tag):
        values = list()
        for node in get_nodes(self, 'uri'):
            values.append(node.text)
        return values

    def _icanhaz_set_list(self, tag, values):
        delattr(self, tag)
        create_nodes(self, tag, values)

    def _icanhaz_del(self, tag):
        for node in get_nodes(self, tag):
            self.remove(node)


class Typable(object):

    @property
    def type(self):
        node = get_node(self, 'type')
        return node.text if node is not None else None

    @type.setter
    def type(self, value):
        delattr(self, 'type')
        create_node(self, 'type', value)

    @type.deleter
    def type(self):
        node = get_node(self, 'type')
        if node is not None:
            self.remove(node)


class Repeatable(object):

    @property
    def min(self):
        return int(self.get('min', 0))

    @min.setter
    def min(self, value):
        if value is None:
            delattr(self, 'min')
            return
        # NOTE: min < 0 or min > max make no sense
        mn = max(int(value), 0)
        if self.max is None:
            self.set('min', str(mn))
        else:
            self.set('min', str(min(mn, self.max)))

    @min.deleter
    def min(self):
        self.attrib.pop('min', 0)

    @property
    def max(self):
        mx = self.get('max')
        return mx if mx is None else int(mx)

    @max.setter
    def max(self, value):
        if value is None:
            delattr(self, 'max')
            return
        # NOTE: max < 1 or max < min make no sense
        self.set('max', str(max(int(value), 1, self.min)))

    @max.deleter
    def max(self):
        self.attrib.pop('max', None)


class Node(object):

    parent = None

    @property
    def children(self):
        return list(self)
        # NOTE: for ALL children, and not just direct ones:
        # return [e for e in self.iter() if e is not self]

    def append(self, child):
        super().append(child)
        try:
            child.parent = self
        except AttributeError:
            pass  # `child` doesn't have this mixin

    def remove(self, child):
        try:
            child.parent = None
        except AttributeError:
            pass  # `child` doesn't have this mixin
        super().remove(child)


__copyright__ = "Copyright the pyHeimdall contributors."
__license__ = 'AGPL-3.0-or-later'
__all__ = [
    'Documentable', 'Typable', 'Repeatable',
    'Node',
    '__copyright__', '__license__',
    ]



# -*- coding: utf-8 -*-
import heimdall
from xml.etree import ElementTree as etree
from urllib.parse import urlparse
from urllib.request import urlopen
from ..decorators import get_database, create_database

"""
Provides connectors to HERA-formatted XML files.

This module defines input and output connectors to databases composed in full or in part of XML files following the HERA schema.

* :py:class:`heimdall.connectors.xml.getDatabase` is the input connector
* :py:class:`heimdall.connectors.xml.serialize` is the output connector

:copyright: The pyHeimdall contributors.
:licence: Afero GPL, see LICENSE for more details.
:SPDX-License-Identifier: AGPL-3.0-or-later
"""  # nopep8: E501


@get_database('hera:xml')
def getDatabase(**options):
    r"""Imports a database from a HERA XML file

    :param \**options: Keyword arguments, see below.
    :Keyword arguments:
        * **url** (:py:class:`str`) -- Local ou remote path of the XML file to import
        * **format** (:py:class:`str`, optional) Always ``hera:xml``
        * **encoding** (:py:class:`str`, optional, default: ``utf-8``) -- ``url`` file encoding
    :return: HERA element tree
    :rtype: :py:class:`xml.etree.ElementTree.Element`

    Usage example: ::

      >>> import heimdall
      >>> tree = heimdall.getDatabase(format='hera:xml', url='some/input.xml')
      >>> # ... do stuff ...

    .. CAUTION::
       For future compability, this function shouldn't be directly called; as shown in the usage example above, it should only be used through :py:class:`heimdall.getDatabase`.
    """  # nopep8: E501
    url = options['url']
    encoding = options.get('encoding', 'utf-8')
    if is_url(url):
        with urlopen(url) as response:
            content = response.read().decode(encoding)
        # can raise urllib.error.HTTPError (HTTP Error 404: Not Found, ...)
    else:
        with open(url, 'r') as f:
            content = f.read()
        # can raise OSError (file not found, ...)
    target = heimdall.elements.Builder()
    parser = etree.XMLParser(target=target)
    return etree.fromstring(content, parser)


def is_url(path):
    schemes = ('http', 'https', 'file', )
    return urlparse(path).scheme in schemes


@create_database('hera:xml')
def serialize(tree, url, pretty=True, **options):
    r"""Serializes a HERA elements tree into an XML file

    :param tree: (:py:class:`xml.etree.ElementTree.Element`) HERA elements tree
    :param url: (:py:class:`str`) Path of the XML file to create
    :param format: (:py:class:`str`) Always ``hera:xml``
    :return: None
    :rtype: :py:class:`NoneType`

    Usage example: ::
      >>> import heimdall
      >>> tree = heimdall.getDatabase(format='hera:xml', url='some/input.xml')
      >>> # ... do stuff ...
      >>> heimdall.serialize(tree, format='hera:xml', url='some/output.xml')
    """
    tree = etree.ElementTree(tree)
    etree.indent(tree, space='  ', level=0)
    with open(url, 'w') as f:
        tree.write(f, encoding='unicode')


__copyright__ = "Copyright the pyHeimdall contributors."
__license__ = 'AGPL-3.0-or-later'
__all__ = [
    'getDatabase',
    'serialize',

    '__copyright__', '__license__',
    ]



# -*- coding: utf-8 -*-

"""
Provides utility functions around HERA elements tree refactoring or cleanup.

:copyright: The pyHeimdall contributors.
:licence: Afero GPL, see LICENSE for more details.
:SPDX-License-Identifier: AGPL-3.0-or-later
"""
from xml.etree.ElementTree import Element, QName


def get_nodes(tree, tag, filter=None, direct=True):
    """Retrieves specific elements in a HERA element tree.

    :param tree: (:py:class:`heimdall.elements.Root`) -- HERA element tree
    :param tag: (:py:class:`str`) -- Tag of elements to search for
    :param filter: (:py:class:`function`, optional) -- Filtering function
    :param direct: (:py:class:`bool`, optional, default: ``True``) -- If ``True``, retrieves ``tree``'s direct children only; if ``False``, perform a deep search
    :return: Found elements
    :rtype: :py:class:`list` of :py:class:`xml.etree.ElementTree.Element`

    This function returns a :py:class:`list` of all elements from ``tree`` of tag ``tag`` and for which ``filter`` returns ``True``. The returned :py:class:`list` is empty if there is no such element.

    :see also: | :py:class:`heimdall.util.get_node` can be used to search for no more than one element in ``tree``.
               | :py:class:`heimdall.getItems`, :py:class:`heimdall.getMetadata`, :py:class:`heimdall.getProperties`, :py:class:`heimdall.getEntities` and :py:class:`heimdall.getAttributes` are dedicated retrieval functions to search for elements in a HERA database, and should be used whenever possible for future compatibility.
    """  # nopep8: E501
    if tree is None:
        return list()
    if direct:
        nodes = tree.findall(f'./{tag}')
    else:
        nodes = tree.findall(f'.//{tag}')
    if filter:
        return [node for node in nodes if filter(node)]
    return nodes


def get_node(tree, tag, filter=None, direct=True):
    """Retrieves a specific element in a HERA element tree.

    :param tree: (:py:class:`heimdall.elements.Root`) -- HERA element tree
    :param tag: (:py:class:`str`) -- Tag of elements to search for
    :param filter: (:py:class:`function`, optional) -- Filtering function
    :param direct: (:py:class:`bool`, optional, default: ``True``) -- If ``True``, retrieves ``tree``'s direct children only; if ``False``, perform a deep search
    :return: Found element
    :rtype: :py:class:`xml.etree.ElementTree.Element`

    This function returns a single element from ``tree`` as defined by ``tag`` and ``filter``, or ``None`` if there is no element in ``tree`` corresponing to ``tag`` and ``filter``.
    It works exactly like :py:class:`heimdall.util.get_nodes`, but raises an ``IndexError`` if it could return more than one result.

    :see also: | :py:class:`heimdall.util.get_nodes` can be used to retrieve multiple elements in ``tree``.
               | :py:class:`heimdall.util.get_root` can retrieve ``tree``'s root element.
               | :py:class:`heimdall.getItem`, :py:class:`heimdall.getProperty`, :py:class:`heimdall.getEntity` and :py:class:`heimdall.getAttribute` are dedicated retrieval functions to search for a specific element in a HERA database, and should be used whenever possible for future compatibility.
    """  # nopep8: E501
    nodes = get_nodes(tree, tag, filter, direct)
    if len(nodes) == 0:
        return None
    if len(nodes) == 1:
        return nodes[0]
    raise IndexError(f"Too many {tag} elements ({len(nodes)})")


def get_root(tree):
    """Retrieves a HERA element tree root.

    :param tree: (:py:class:`heimdall.elements.Root`) -- HERA element tree
    :return: Tree root
    :rtype: :py:class:`xml.etree.ElementTree.Element`

    :see also: :py:class:`heimdall.util.get_node` can be used to search for any element in ``tree``.
    """  # nopep8: E501
    return tree.xpath('//hera')[0]


def get_language(node):
    """Retrieves the language of a node.

    :param node: (:py:class:`xml.etree.ElementTree.Element`) -- HERA element
    :return: ``node`` language
    :rtype: :py:class:`str`

    This function returns ``node``'s language, if it has any.
    The returned language `should be a language code <https://www.w3.org/International/questions/qa-when-xmllang.en.html>`_, but it will be any :py:class:`str` that was set in ``node``'s ``xml:lang`` attribute.

      >>> import heimdall
      >>> # ...
      >>> metadata = heimdall.getMetadata(item, ...)
      >>> language_code = heimdall.util.get_language(metadata)
      >>> # if metadata has: xml:lang='en_US',
      >>> #   then language_code would be "en_US".
      >>> # if metadata has: xml:lang='lolspeak',
      >>> #   then language_code would be "lolspeak".
      >>> # if metadata didn't have any `xml:lang` set,
      >>> #   then language_code would be `None`.

    :see also: :py:class:`heimdall.util.set_language` to set ``node``'s language.
    """  # nopep8: E501
    qname = QName('http://www.w3.org/XML/1998/namespace', 'lang')
    return node.get(qname)


def set_language(node, language):
    """Sets the language of a node.

    :param node: (:py:class:`xml.etree.ElementTree.Element`) -- HERA element
    :param language: (:py:class:`str`) -- Language code (`see documentation <https://www.w3.org/International/questions/qa-when-xmllang.en.html>`_
    :return: None
    :rtype: :py:class:`NoneType`

    This function set ``node``'s ``xml:lang`` attribute to value ``language``.

      >>> import heimdall
      >>> # ...
      >>> metadata = heimdall.getMetadata(item, ...)
      >>> heimdall.util.set_language(metadata, 'fr')
      >>> # now metadata has: xml:lang='fr'.
      >>> heimdall.util.set_language(metadata, None)
      >>> # now metadata has no more xml:lang set.

    :see also: :py:class:`heimdall.util.get_language` to retrieve ``node``'s language.
    """  # nopep8: E501
    qname = QName('http://www.w3.org/XML/1998/namespace', 'lang')
    return node.set(qname, language)


def create_nodes(parent, tag, text):
    r"""Creates a child, or children, for a node.

    :param parent: (:py:class:`xml.etree.ElementTree.Element`) -- HERA element
    :param tag: (:py:class:`str`) -- Created child tag, or tag common to all created children
    :param tag: (:py:class:`str`\ \|\ :py:class:`dict`) -- See description
    :return: Children created
    :rtype: :py:class:`list` of :py:class:`xml.etree.ElementTree.Element`

    This function can be used to append a single child to a ``parent`` node: ::

      >>> import heimdall
      >>> # ...
      >>> entity = heimdall.getEntity(tree, ...)
      >>> heimdall.util.create_nodes(entity, 'description', "blah blah blah")

    This function can be used to append in one go multiple localized nodes to ``parent``: ::

      >>> import heimdall
      >>> # ...
      >>> entity = heimdall.getEntity(tree, ...)
      >>> heimdall.util.create_nodes(entity, 'description', {
      >>>     'de': "blabla",
      >>>     'en': "blah blah blah",
      >>>     'fr': "bla-bla",
      >>>     })

    :see also: :py:class:`heimdall.util.create_node` if you don't do internationalization.
    """  # nopep8: E501
    nodes = list()

    def _create(_parent, _tag, _text):
        _node = create_node(_parent, _tag, _text)
        nodes.append(_node)
        return _node

    if type(text) is str or type(text) is type(None):
        _create(parent, tag, text)
    elif type(text) is dict:
        for language_key, value in text.items():
            if type(value) is str or type(value) is type(None):
                node = _create(parent, tag, value)
                set_language(node, language_key)
            else:
                for v in value:
                    node = _create(parent, tag, v)
                    set_language(node, language_key)
    else:  # type(text) is list
        for value in text:
            _create(parent, tag, value)
    return nodes


def create_node(parent, tag, text=None):
    """Creates a child for a node.

    :param parent: (:py:class:`xml.etree.ElementTree.Element`) -- HERA element
    :param tag: (:py:class:`str`) -- Created child tag
    :param tag: (:py:class:`str`) -- Created child text
    :return: Child
    :rtype: :py:class:`xml.etree.ElementTree.Element`

    This function appends a single child to a ``parent`` node: ::

      >>> import heimdall
      >>> # ...
      >>> entity = heimdall.getEntity(tree, ...)
      >>> heimdall.util.create_nodes(entity, 'description', "blah blah blah")

    :see also: :py:class:`heimdall.util.create_nodes` can be used to append in one go multiple localized nodes to ``parent``.
    """  # nopep8: E501
    if type(text) is not str and type(text) is not type(None):
        bad = str(type(text).__name__)
        raise TypeError(f"Argument 'text' must be str, got '{bad}'")

    node = Element(tag)
    parent.append(node)
    if text is not None:
        node.text = text
    return node


def get_container(tree, tag):
    nodes = tree.findall(f'.//{tag}')
    if len(nodes) == 0:
        return None
    assert len(nodes) == 1
    return nodes[0]


def create_container(tree, tag):
    container = get_container(tree, tag)
    if container is None:
        container = Element(tag)
        tree.append(container)
    return container


def update_node_value(node, key, value):
    if value is not None:
        node.set(key, str(value))
    else:  # remove value
        node.attrib.pop(key)


def maybe_update_node_values(node, keys, **kwargs):
    for key in keys:
        try:
            value = kwargs[key]
            update_node_value(node, key, value)
        except KeyError:
            pass  # nothing to change


def update_node_children(node, key, value):
    children = get_nodes(node, key)
    if len(children) == 0:
        create_nodes(node, key, value)
    for child in children:
        if value is not None:
            child.text = str(value)
        else:  # remove child node
            node.remove(child)


def maybe_update_node_children(node, keys, **kwargs):
    for key in keys:
        try:
            value = kwargs[key]
            update_node_children(node, key, value)
        except KeyError:
            pass  # nothing to change


__copyright__ = "Copyright the pyHeimdall contributors."
__license__ = 'AGPL-3.0-or-later'
__all__ = [
    'get_node', 'get_nodes', 'get_root',
    'get_language', 'set_language',
    'create_node', 'create_nodes',
    '__copyright__', '__license__',
    ]





"""
Provides CRUD operations to search for or
edit metadata in a HERA item.

:copyright: The pyHeimdall contributors.
:licence: Afero GPL, see LICENSE for more details.
:SPDX-License-Identifier: AGPL-3.0-or-later
"""

import heimdall
from .util import (
        get_node as _get_node,
        create_nodes as _create_nodes,
    )


def getMetadata(item, **options):
    r"""Retrieves metadata elements from an item.

    :param item: (:py:class:`heimdall.elements.Item`) -- HERA item
    :param \**options: (:py:class:`dict`, optional) -- Keyword arguments, see description.
    :return: Metadata list
    :rtype: :py:class:`list` of :py:class:`heimdall.elements.Metadata`

    Usage examples: ::

      >>> import heimdall
      >>> # ... do stuff ...
      >>> item = getItem(...)
      >>> # consider we just got the following item:
      >>> # <item eid='person'>
      >>> #   <metadata pid='dc:title' xml:lang='en'>William</metadata>
      >>> #   <metadata pid='dc:title' xml:lang='fr'>Guillaume</metadata>
      >>> #   <metadata pid='pet' xml:lang='en'>Chirpy</metadata>
      >>> #   <metadata pid='pet' xml:lang='en'>Claws</metadata>
      >>> #   <metadata pid='pet'>Blackie</metadata>
      >>> # </item>
      >>> # then we'll have the following results
      >>> metadata = heimdall.getMetadata(item, pid='dc:title')
      >>> # len(metadata) == 2  (metadata of values William, Guillaume)
      >>> metadata = heimdall.getMetadata(item, pid='pet')
      >>> # len(metadata) == 3  (metadata of values Chirpy, Claws, Blackie)
      >>> metadata = heimdall.getMetadata(item, pid='meats')
      >>> # len(metadata) == 0  (no metadata in item with this pid)

    This function can be used to retrieve metadata in a specific language,
    using the ``lang`` parameter: ::

      >>> import heimdall
      >>> # ... do stuff ...
      >>> item = getItem(...)
      >>> # consider we just got the following item:
      >>> # <item eid='person'>
      >>> #   <metadata pid='dc:title' xml:lang='en'>William</metadata>
      >>> #   <metadata pid='dc:title' xml:lang='fr'>Guillaume</metadata>
      >>> #   <metadata pid='pet' xml:lang='en'>Chirpy</metadata>
      >>> #   <metadata pid='pet' xml:lang='en'>Claws</metadata>
      >>> #   <metadata pid='pet'>Blackie</metadata>
      >>> # </item>
      >>> # then we'll have the following results
      >>> metadata = heimdall.getMetadata(item, lang='en')
      >>> # len(metadata) == 3  (metadata of values William, Chirpy, Claws)
      >>> metadata = heimdall.getMetadata(item, lang='fr')
      >>> # len(metadata) == 1  (metadata of value Guillaume)
      >>> metadata = heimdall.getMetadata(item, lang=None)
      >>> # len(metadata) == 1  (metadata of value Blackie is not localized)
      >>> metadata = heimdall.getMetadata(item, lang='ar')
      >>> # len(metadata) == 0  (no metadata localized in arabic in item)
      >>> metadata = heimdall.getMetadata(item, pid='pet', lang='ar')
      >>> # len(metadata) == 0  (no metadata 'pet' localized in arabic in item)
      >>> metadata = heimdall.getMetadata(item, pid='dc:title', lang='fr')
      >>> # len(metadata) == 3  (metadata of value Guillaume)
      >>> metadata = heimdall.getMetadata(item, pid='pet', lang='fr')
      >>> # len(metadata) == 0  (no metadata 'pet' is localized in french)
      >>> metadata = heimdall.getMetadata(item, pid='pet', lang='en')
      >>> # len(metadata) == 2  (metadata of values Chirpy, Claws)

    Parameter ``options`` can be anything you want to filter metadata with,
    even if you target something outside of the scope of the HERA specification: ::

      >>> import heimdall
      >>> # ... do stuff ...
      >>> item = getItem(...)
      >>> # consider we just got the following item:
      >>> # <item eid='person'>
      >>> #   <metadata pid='dc:title' xml:lang='en'>William</metadata>
      >>> #   <metadata pid='dc:title' xml:lang='fr' floofs='mild'>Guillaume</metadata>
      >>> #   <metadata pid='pet' xml:lang='en' floofs='high'>Chirpy</metadata>
      >>> #   <metadata pid='pet' xml:lang='en' floofs='mild'>Claws</metadata>
      >>> #   <metadata pid='pet' floofs='high'>Blackie</metadata>
      >>> # </item>
      >>> # then we'll have the following results
      >>> metadata = heimdall.getMetadata(item, floofs='high')
      >>> # len(metadata) == 2  (metadata of values Chirpy, Blackie)
      >>> metadata = heimdall.getMetadata(item, floofs='mild')
      >>> # len(metadata) == 2  (metadata of values Guillaume, Claws)
      >>> metadata = heimdall.getMetadata(item, floofs='high', lang='en')
      >>> # len(metadata) == 1  (metadata of value Chirpy)
      >>> metadata = heimdall.getMetadata(item, floofs=None)
      >>> # len(metadata) == 1  (metadata of values William, Guillaume, Blackie)

    If ``options`` are left undefined, ``getMetadata`` simply returns all metadata elements of an item.
    In the above examples, ``heimdall.getMetadata(item)`` would return a list of all (5) metadata elements of ``item``.

    To get results directly as :py:class:`str` values instead of as metadata elements, :py:class:`heimdall.getValues` can be used with a similar behavior.
    """  # nopep8: E501
    filters = dict()
    for key, wanted in options.items():
        if key == 'lang':
            filters[key] = (wanted, lambda m, _: heimdall.util.get_language(m))
        else:
            filters[key] = (wanted, lambda m, param: m.get(param))
    metadata = item.metadata
    for param, (wanted, _get) in filters.items():
        metadata = [m for m in metadata if _get(m, param) == wanted]
    return metadata


def getValues(item, **options):
    r"""Retrieves metadata values from an item.

    :param item: (:py:class:`heimdall.elements.Item`) -- HERA item
    :param \**options: (:py:class:`dict`, optional) -- Keyword arguments, see description.
    :return: Metadata values list
    :rtype: :py:class:`list` of :py:class:`str`

    This function can be used to retrieve metadata values from an item ``item``.
    It behaves exactly like :py:class:`heimdall.getMetadata`, but returns directly metadata values as a :py:class:`list` of :py:class:`str`, instead of metadata as a :py:class:`list` of :py:class:`heimdall.elements.Metadata`.

    Please note that if some of ``item``'s metadata value are empty, this function returns empty strings, not :py:class:`NoneType` elements: ::

      >>> import heimdall
      >>> # ... do stuff ...
      >>> item = getItem(...)
      >>> # consider we just got the following item:
      >>> # <item>
      >>> #   <metadata>Donald</metadata>
      >>> #   <metadata aid='nephew'>Huey</metadata>
      >>> #   <metadata aid='nephew'></metadata>
      >>> #   <metadata aid='nephew'>Louie</metadata>
      >>> # </item>
      >>> # then we'll have the following results
      >>> values = heimdall.getValues(item)
      >>> # values = ["Donald", "Huey", "", "Louie"]
      >>> values = heimdall.getValues(item, aid='nephew')
      >>> # values = ["Huey", "", "Louie"]
    """  # nopep8: E501
    metadata = heimdall.getMetadata(item, **options)
    values = list()
    for m in metadata:
        if m.text is not None:
            values.append(m.text)
        else:
            values.append('')
    return values


def getValue(item, **options):
    r"""Retrieves a single metadata value from an item.

    :param item: (:py:class:`heimdall.elements.Item`) -- HERA item
    :param \**options: (:py:class:`dict`, optional) -- Keyword arguments, see description.
    :return: Metadata value
    :rtype: :py:class:`str`

    This function works exactly like :py:class:`heimdall.getValues`, but raises an :py:class:`IndexError` if there is more than one metadata corresponding to parameters.
    """  # nopep8: E501
    values = getValues(item, **options)
    if len(values) == 0:
        return None
    if len(values) == 1:
        return values[0]
    raise IndexError("Too many metadata ({count})".format(count=len(values)))


def createMetadata(item, value, **options):
    r"""Adds a single metadata to an item.

    :param item: (:py:class:`heimdall.elements.Item`) -- HERA item
    :param value: (:py:class:`str`\ \|\ :py:class:`dict`) -- Metadata value
    :param \**options: (:py:class:`dict`, optional) -- Keyword arguments, see description.
    :return: Metadata list
    :rtype: :py:class:`list` of :py:class:`heimdall.elements.Metadata`

    Metadata created by ``createMetadata`` will always be added to the item ``item``, with no consistency check.
    For example, ``createMetadata`` does not verify that ``pid`` is a valid property identifier in the database ``item`` comes from.
    Should ``item`` have a related entity, ``createMetadata`` does not check that the new metadata makes sense for the entity (*ie.* that the entity defines an attribute reusing the property ``pid``).

    The following example adds a new metadata to an existing item: ::

      >>> import heimdall
      >>> # ... do stuff ...
      >>> item = getItem(...)
      >>> heimdall.createMetadata(item, 'Bill', pid='dc:title', aid='name')
      >>> # we just added the following metadata to `item`:
      >>> # <metadata aid='name' pid='dc:title'>Bill</metadata>

    Metadata added to ``item`` can be localized, if their value is given as a :py:class:`dict` instead of a :py:class:`str` (the :py:class:`dict` keys should be valid language codes).
    Here is an example: ::

      >>> import heimdall
      >>> # ... do stuff ...
      >>> item = getItem(...)
      >>> heimdall.createMetadata(item, {
      >>>     'en_AU': 'Bill',
      >>>     'de_DE': 'Wilhelm',
      >>>     },
      >>>     pid='dc:title',
      >>>     aid='name')
      >>> # we just added the following 2 metadata to `item`:
      >>> # <metadata aid='name' pid='dc:title' xml:lang='en_AU'>Bill</metadata>
      >>> # <metadata aid='name' pid='dc:title' xml:lang='de_DE'>Wilhelm</metadata>
    """  # nopep8: E501
    nodes = _create_nodes(item, 'metadata', value)
    for node in nodes:
        for key, param in options.items():
            node.set(key, param)
    return nodes


def deleteMetadata(item, **options):
    """TODO
    """
    nodes = getMetadata(item, **options)
    for node in nodes:
        item.remove(node)


__copyright__ = "Copyright the pyHeimdall contributors."
__license__ = 'AGPL-3.0-or-later'
__all__ = [
    'getMetadata', 'getValue', 'getValues',
    'createMetadata', 'deleteMetadata',
    '__copyright__', '__license__',
    ]



# -*- coding: utf-8 -*-

"""
Provides CRUD operations to search for or
edit attributes in a HERA entity.

:copyright: The pyHeimdall contributors.
:licence: Afero GPL, see LICENSE for more details.
:SPDX-License-Identifier: AGPL-3.0-or-later
"""

from .util.tree import (
        get_node as _get_node,
        get_nodes as _get_nodes,
        get_root as _get_root,
        create_node as _create_node,
        create_nodes as _create_nodes,
        maybe_update_node_values as _maybe_update_values,
        maybe_update_node_children as _maybe_update_children,
    )
from .elements import Attribute


def getAttribute(entity, filter):
    """Retrieves a single attribute from an entity.

    :param entity: (:py:class:`heimdall.elements.Entity`) -- HERA entity
    :param filter: (:py:class:`function`, optional) -- Filtering function
    :return: Attribute from ``entity``
    :rtype: :py:class:`heimdall.elements.Attribute`

    This function works exactly like :py:class:`heimdall.getAttributes`, but raises an ``IndexError`` if ``filter`` returns more than one result.
    """  # nopep8: E501
    return _get_node(entity, 'attribute', filter)


def getAttributes(entity, filter=None):
    """Retrieves attributes from an entity.

    :param entity: (:py:class:`heimdall.elements.Entity`) -- HERA entity
    :param filter: (:py:class:`function`, optional) -- Filtering function
    :return: Attributes subset from ``entity``
    :rtype: :py:class:`list` of :py:class:`heimdall.elements.Attribute`

    This function can be used to retrieve all attributes of an entity: ::

      >>> import heimdall
      >>> ...  # create or load HERA tree
      >>> entity = heimdall.getEntity(tree, ...)  # retrieve an entity
      >>> attributes = heimdall.getAttributes(enitity)  # retrieve all attributes

    To retrieve only *some* attributes, you can use a filter.
    A filter is a function which takes only one attribute as a parameter, and returns either ``True`` (we want this attribute to be part of the list returned by ``getAttributes``) or ``False`` (we don't want it). ::

      >>> import heimdall
      >>> ...  # create or load HERA tree
      >>> entity = heimdall.getEntity(tree, ...)  # retrieve an entity
      >>> def by_pid(attribute):  # create filtering function
      >>>     return attribute.get('pid') == 'dc:title'
      >>> # retrieve only attributes equivalent to a title
      >>> names = heimdall.getAttributes(entity, by_pid)

    For simple filters, anonymous functions can of course be used: ::

      >>> import heimdall
      >>> ...  # create or load HERA tree
      >>> entity = heimdall.getEntity(tree, ...)  # retrieve an entity
      >>> # retrieve only attributes equivalent to an identifier
      >>> ids = heimdall.getAttributes(entity, lambda a: a.get('pid') == 'id')
    """  # nopep8: E501
    return _get_nodes(entity, 'attribute', filter)


def createAttribute(entity, pid, **kwargs):
    r"""Adds a single attribute to an entity.

    :param entity: (:py:class:`heimdall.elements.Entity`) -- HERA entity
    :param pid: (:py:class:`str`) -- Linked property ID
    :param \**kwargs: (:py:class:`dict`, optional) -- Keyword arguments, see below.
    :Keyword arguments:
        * **id** (:py:class:`str`, optional) -- Attribute id
        * **min** (:py:class:`int`, optional, default: ``0``) -- Minimum occurences in an item: ``1`` or more means it is mandatory
        * **max** (:py:class:`int`, optional) -- Maximum occurences in an item: ``1`` means it is not repeatable, more (or no value) means it is; ``0`` or ``<min`` is undefined behaviour
        * **type** (:py:class:`str`, optional) -- Type override; no value means the linked property ``type`` will be used
        * **name** (:py:class:`str`\ \|\ :py:class:`dict`, optional) -- Name override; no value means the linked property ``name`` will be used
        * **description** (:py:class:`str`\ \|\ :py:class:`dict`, optional) -- Description override; no value means the linked property ``description`` will be used
        * **uri** (:py:class:`list` of :py:class:`str`, optional) -- URI list override; no value means the linked property ``uri`` will be used, any value adds to the property ``uri``
    :return: Attribute that was just created
    :rtype: :py:class:`heimdall.elements.Attribute`

    This function can be used to add a new attribute to an existing entity.
    Attributes created by ``createAttribute`` will always be added to the
    entity ``entity``, with no consistency check.
    For example, ``createAttribute`` does not verify that ``pid`` is a valid
    property identifier in the database ``entity`` comes from.

    In its simplest usage, ``createAttribute`` simply creates
    a new ``<attribute pid='xxx'/>`` element: ::

      >>> import heimdall
      >>> ...
      >>> entity = heimdall.getEntity(...)
      >>> heimdall.createAttribute(entity, pid='xxx')  # create a new attribute
      >>> # the following child is now added to entity attributes list:
      >>> # <attribute pid='xxx' />

    Additional supported parameters are ``type``, ``name``, ``description``,
    and ``uri``.
    Each of these parameters creates appropriate children for the attribute.
    Here is an example: ::

      >>> import heimdall
      >>> ...
      >>> heimdall.createAttribute(entity,
      >>>     pid='dc:title', name='Label',
      >>>     uri=[
      >>>         'http://purl.org/dc/terms/title',
      >>>         'http://purl.org/dc/terms/alternative',
      >>>         'http://nakala.fr/terms#title'
      >>>     ])
      >>> # the following attribute is now added to the entity:
      >>> # <attribute pid='dc:title'>
      >>> #     <type>text</type>
      >>> #     <name>Label</name>
      >>> #     <uri>http://purl.org/dc/terms/title</uri>
      >>> #     <uri>http://purl.org/dc/terms/alternative</uri>
      >>> #     <uri>http://nakala.fr/terms#title</uri>
      >>> # </attribute>

    Please note that ``name`` and ``description`` can be localized, if they are
    of type :py:class:`dict` instead of :py:class:`str` (:py:class:`dict` keys are language codes).
    The following example shows attribute localization: ::

      >>> import heimdall
      >>> ...  # create HERA element tree
      >>> heimdall.createAttribute(entity,
      >>>      pid='dc:title',
      >>>      max='1',
      >>>      name={
      >>>          'de': "Name",
      >>>          'en': "Name",
      >>>          'fr': "Nom",
      >>>      },
      >>>      description={
      >>>          'en': "Human-readable name",
      >>>          'fr': "Nom usuel",
      >>>      })
      >>> # the following element is now added to the corresponding ``<entity/>`` element:
      >>> # <attribute pid='dc:title' min='0' max='1'>
      >>> #     <type>text</type>
      >>> #     <name xml:lang='de'>Name</name>
      >>> #     <name xml:lang='en'>Name</name>
      >>> #     <name xml:lang='fr'>Nom</name>
      >>> #     <description xml:lang='en'>Human-readable name</description>
      >>> #     <description xml:lang='fr'>Nom usuel</description>
      >>> # </attribute>
    """  # nopep8: E501
    param = kwargs.get('id', None)
    if param is not None:
        # Check attribute unique id (aid) does not exist
        if getAttribute(entity, lambda a: a.get('id') == param) is not None:
            raise ValueError(f"Attribute id '{param}' already exists")
        param = str(param)
    node = Attribute(pid=pid)
    if param is not None:
        node.set('id', param)
    param = kwargs.get('min', 0)
    node.min = param
    param = kwargs.get('max', None)
    node.max = param

    param = kwargs.get('type', None)
    if param is not None:
        _create_node(node, 'type', param)
    param = kwargs.get('name', None)
    if param is not None:
        node.name = param
    param = kwargs.get('description', None)
    if param is not None:
        node.description = param
    param = kwargs.get('uri', list())
    if type(param) is not list:
        raise TypeError(f"'uri' expected:'list', got:'{type(param).__name__}'")
    node.uri = param

    entity.append(node)
    return node


def replaceAttribute(attribute, **kwargs):
    """Replaces an existing attribute.

    .. ERROR::
       This feature is not implemented yet.
       This can be either due to lack of resources, lack of demand, because it wouldn't be easily maintenable, or any combination of these factors.

       Interested readers can submit a request to further this topic.
       See the ``CONTRIBUTING.rst`` file at the root of the repository for details.
    """  # nopep8: E501
    raise ValueError("TODO: Not Implemented")


def updateAttribute(attribute, **kwargs):
    r"""Updates a single attribute.

    :param attribute: (:py:class:`heimdall.elements.Attribute`) -- HERA attribute
    :param \**kwargs: (:py:class:`dict`, optional) -- Keyword arguments, see below.
    :Keyword arguments:
        * **id** (:py:class:`str`, optional) -- Attribute id
        * **pid** (:py:class:`str`, optional) -- Property id
        * **min** (:py:class:`int`, optional, default: ``0``) -- Minimum occurences in an item: ``1`` or more means it is mandatory
        * **max** (:py:class:`int`, optional) -- Maximum occurences in an item: ``1`` means it is not repeatable, more (or no value) means it is; ``0`` or ``<min`` is undefined behaviour
        * **type** (:py:class:`str`, optional) -- Type override; no value means the linked property ``type`` will be used
        * **name** (:py:class:`str`\ \|\ :py:class:`dict`, optional) -- Name override; no value means the linked property ``name`` will be used
        * **description** (:py:class:`str`\ \|\ :py:class:`dict`, optional) -- Description override; no value means the linked property ``description`` will be used
        * **uri** (:py:class:`list` of :py:class:`str`, optional) -- URI list override; no value means the linked property ``uri`` will be used, any value adds to the property ``uri``
    :return: Attribute that was just updated
    :rtype: :py:class:`heimdall.elements.Attribute`

    This function can be used to update an existing attribute.
    Like ``createAttribute``, it makes with no consistency check.
    For example, ``createAttribute`` does not verify that ``pid`` is a valid property identifier in the database ``attribute`` comes from.

    Any or all of ``updateAttribute`` parameters can be used to update ``attribute``: ::

      >>> import heimdall
      >>> # ...
      >>> attribute = heimdall.getAttribute(entity, ...)
      >>> heimdall.updateAttribute(attribute, name='Label', min=1)

    Please note that ``name`` and ``description`` can be localized, if they are of type :py:class:`dict` instead of :py:class:`str` (:py:class:`dict` keys are language codes).
    The following example shows attribute localization: ::

      >>> import heimdall
      >>> # ...
      >>> attribute = heimdall.getAttribute(entity, ...)
      >>> heimdall.updateAttribute(attribute,
      >>>      name={
      >>>          'de': "Name",
      >>>          'en': "Name",
      >>>          'fr': "Nom",
      >>>      },
      >>>      description={
      >>>          'en': "Human-readable name",
      >>>          'fr': "Nom usuel",
      >>>      })
    """  # nopep8: E501
    VALUES = ['id', 'pid', 'min', 'max', ]
    CHILDREN = ['type', 'name', 'description', 'uri', ]
    _maybe_update_values(attribute, VALUES, **kwargs)
    _maybe_update_children(attribute, CHILDREN, **kwargs)
    return attribute


def deleteAttribute(entity, filter):
    """Deletes a single Attribute from an entity.

    :param entity: (:py:class:`heimdall.elements.Entity`) -- HERA entity
    :param filter: (:py:class:`function`) -- Filtering function
    :return: None
    :rtype: :py:class:`NoneType`

    This function can be used to remove attributes from an entity.
    It performs the attribute deletion "in place" ; in other words, parameter ``tree`` is directly modified, and this function returns nothing.

    If ``filter`` returns no result, this function does nothing, and does not raise any error.

    This function doesn't delete any metadata referencing the ``pid`` of the deleted attribute.

    Usage ::

      >>> import heimdall
      >>> ...  # create config, load HERA tree
      >>> # delete an attribute of a specific entity
      >>> e = heimdall.getEntity(tree, lambda e : e.get('id') == 'person')
      >>> heimdall.deleteAttribute(e, lambda a: a.get('pid') == 'religion')
    """  # nopep8: E501
    nodes = getAttributes(entity, filter)
    for node in nodes:
        entity.remove(node)


__copyright__ = "Copyright the pyHeimdall contributors."
__license__ = 'AGPL-3.0-or-later'
__all__ = [
    'getAttribute', 'getAttributes',
    'createAttribute', 'deleteAttribute',
    'replaceAttribute', 'updateAttribute',
    '__copyright__', '__license__',
    ]



# -*- coding: utf-8 -*-
import heimdall
from heimdall.util import get_node, get_nodes, get_language
from ..decorators import get_database, create_database
from json import load, loads, dump
from urllib.parse import urlparse
from urllib.request import urlopen
from xml.etree.ElementTree import Element

"""
Provides connectors to HERA-formatted JSON files.

This module defines input and output connectors to databases composed in full or in part of JSON files following the HERA schema.

* :py:class:`heimdall.connectors.json.getDatabase` is the input connector
* :py:class:`heimdall.connectors.json.serialize` is the output connector

:copyright: The pyHeimdall contributors.
:licence: Afero GPL, see LICENSE for more details.
:SPDX-License-Identifier: AGPL-3.0-or-later
"""  # nopep8: E501


@get_database('hera:json')
def getDatabase(**options):
    r"""Imports a database from a HERA JSON file

    :param \**options: Keyword arguments, see below.
    :Keyword arguments:
        * **url** (:py:class:`str`) -- Local ou remote path of the JSON file to import
        * **format** (:py:class:`str`, optional) Always ``hera:json``
        * **encoding** (:py:class:`str`, optional, default: ``utf-8``) -- ``url`` file encoding
    :return: HERA element tree
    :rtype: :py:class:`xml.etree.ElementTree.Element`

    Usage example: ::

      >>> import heimdall
      >>> tree = heimdall.getDatabase(format='hera:json', url='some/input.json')
      >>> # ... do stuff ...

    .. CAUTION::
       For future compability, this function shouldn't be directly called; as shown in the usage example above, it should only be used through :py:class:`heimdall.getDatabase`.
    """  # nopep8: E501
    url = options['url']
    encoding = options.get('encoding', 'utf-8')
    if not is_url(url):
        with open(url, 'r') as f:
            data = load(f)
    else:
        with urlopen(url) as response:
            data = loads(response.read().decode(encoding))
    return _create_tree(data)


def is_url(path):
    schemes = ('http', 'https', )
    return urlparse(path).scheme in schemes


def _create_tree(data):
    root = heimdall.createDatabase()

    # create Properties if any
    properties = data.get('properties', None)
    if properties is not None:
        elements = root.get_container('properties')
        for o in properties:
            heimdall.createProperty(root, **o)

    # create Entities if any
    entities = data.get('entities', None)
    if entities is not None:
        elements = root.get_container('entities')
        for o in entities:
            e = heimdall.createEntity(root, **o)
            attributes = o.get('attributes', list())
            for c in attributes:
                heimdall.createAttribute(e, **c)

    # create Items if any
    items = data.get('items', None)
    if items is not None:
        elements = root.get_container('items')
        for o in items:
            metadata = o.pop('metadata', list())
            e = heimdall.createItem(root, **o)
            for m in metadata:
                value = m.pop('value', None)
                heimdall.createMetadata(e, value, **m)

    return root


@create_database('hera:json')
def serialize(tree, url, **options):
    r"""Serializes a HERA elements tree into a JSON file

    :param tree: HERA elements tree
    :param url: Path of the JSON file to create

    .. ERROR::
       This feature is not implemented yet.
       This can be either due to lack of resources, lack of demand, because it
       wouldn't be easily maintenable, or any combination of these factors.

       Interested readers can submit a request to further this topic.
       See the ``CONTRIBUTING`` file at the root of the repository for details.
    """
    data = _tree2object(tree)
    # write data to file
    indent = options.get('indent', 2)
    protek = options.get('ensure_ascii', False)
    with open(url, 'w', encoding='utf-8') as f:
        dump(data, f, indent=indent, ensure_ascii=protek)


def _tree2object(tree):
    data = dict()
    # convert properties to json-ready objects, add them to data
    elements = heimdall.getProperties(tree)
    if len(elements) > 0:
        data['properties'] = list()
    for element in elements:
        data['properties'].append(_property2object(element))
    # convert entities to json-ready objects, add them to data
    elements = heimdall.getEntities(tree)
    if len(elements) > 0:
        data['entities'] = list()
    for element in elements:
        data['entities'].append(_entity2object(element))
    # convert items to json-ready objects, add them to data
    elements = heimdall.getItems(tree)
    if len(elements) > 0:
        data['items'] = list()
    for element in elements:
        data['items'].append(_item2object(element))

    return data


def _property2object(p):
    return _xml2object(p)


def _entity2object(e):
    data = _xml2object(e)
    elements = heimdall.getAttributes(e)
    if len(elements) > 0:
        data['attributes'] = list()
    for element in elements:
        data['attributes'].append(_attribute2object(element))
    return data


def _attribute2object(a):
    data = _xml2object(a)
    data['pid'] = a.get('pid')
    data['min'] = a.get('min', 0)
    _param2text(a, data, 'max', None)
    return data


def _xml2object(x):
    data = dict()
    value = x.get('id', None)
    if value is not None:
        data['id'] = value
    # child --> unique value
    for tag in ['type', ]:
        try:
            array_of_one = [c.text for c in x.children if c.tag == tag]
            assert len(array_of_one) < 2
            if len(array_of_one) > 0 and array_of_one[0] is not None:
                data[tag] = array_of_one[0]
        except AttributeError:
            pass  # no type
    # child --> localized object
    for tag in ['name', 'description', ]:
        obj = _nodes2object(get_nodes(x, tag))
        if obj is not None:
            data[tag] = obj
    # child --> repeatable value
    for tag in ['uri', ]:
        array = _nodes2array(get_nodes(x, tag))
        if len(array) > 0:
            data[tag] = array
    return data


def _item2object(i):
    data = dict()
    for key in ['eid', ]:
        _param2text(i, data, key)
    elements = heimdall.getMetadata(i)
    if len(elements) > 0:
        data['metadata'] = list()
    for element in elements:
        data['metadata'].append(_metadata2object(element))
    return data


def _metadata2object(m):
    data = dict()
    for key in ['pid', 'aid', ]:
        _param2text(m, data, key)
    value = get_language(m)
    if value is not None:
        data['language'] = value
    value = _node2text(m)
    if value is not None:
        data['value'] = value
    return data


def _param2text(x, data, key, default=None):
    value = x.get(key, default)
    if value is not None:
        data[key] = value


def _node2text(node):
    return node.text


def _nodes2object(nodes):
    data = dict()
    for node in nodes:
        language = get_language(node)
        if language is not None:
            data[language] = node.text
        else:
            # emergency exit:
            # if there is even only 1 node with no language,
            # we'll return this one only
            return node.text
    if len(data) < 1:
        return None
    return data


def _nodes2array(nodes):
    data = list()
    for node in nodes:
        if node.text is not None:
            data.append(node.text)
    return data



# -*- coding: utf-8 -*-

"""
Provides CRUD operations to search for or
edit properties in a HERA element tree.

:copyright: The pyHeimdall contributors.
:licence: Afero GPL, see LICENSE for more details.
:SPDX-License-Identifier: AGPL-3.0-or-later
"""

from heimdall.util.tree import (
        get_node, get_nodes,
        get_container, create_container,
        maybe_update_node_values as _maybe_update_values,
        maybe_update_node_children as _maybe_update_children,
    )
from .elements import Property


def getProperty(tree, filter):
    """Retrieves a single property.

    :param tree: (:py:class:`heimdall.elements.Root`) -- HERA element tree
    :param filter: (:py:class:`function`) -- Filtering function
    :return: Property element
    :rtype: :py:class:`heimdall.elements.Property`

    This function returns a single property from ``tree`` as defined by ``filter``, or ``None`` if there is no property in ``tree`` corresponing to ``filter`` (or if ``tree`` has no property).
    It works exactly like :py:class:`heimdall.getProperties`, but raises an ``IndexError`` if ``filter`` returns more than one result.
    """  # nopep8: E501
    container = get_node(tree, 'properties')
    return get_node(container, 'property', filter)


def getProperties(tree, filter=None):
    """Retrieves a collection of properties.

    :param tree: (:py:class:`heimdall.elements.Root`) -- HERA element tree
    :param filter: (:py:class:`function`, optional) -- Filtering function
    :return: List of Property elements
    :rtype: :py:class:`list` of :py:class:`heimdall.elements.Property`

    This function can be used to retrieve all properties in a database: ::

      >>> import heimdall
      >>> ...  # create config
      >>> tree = heimdall.getDatabase(config)  # load HERA tree
      >>> properties = heimdall.getProperties(tree)  # retrieve all properties

    To retrieve only *some* properties, you can use a filter.
    A filter is a function which takes only a property as a parameter, and returns either ``True`` (we want this property to be part of the list returned by ``getProperties``) or ``False`` (we don't want it). ::

      >>> import heimdall
      >>> ...  # create config, load HERA tree
      >>> def by_pid(property):  # create filtering function
      >>>     return property.get('id') == 'dc:title'
      >>> # retrieve only a specific property by its id
      >>> properties = heimdall.getProperties(tree, by_pid)

    For simple filters, anonymous functions can of course be used: ::

      >>> import heimdall
      >>> ...  # create config, load HERA tree
      >>> # retrieve one specific property by its id
      >>> heimdall.getProperties(tree, lambda e: e.get('id') == 'dc:title')
    """  # nopep8: E501
    container = get_node(tree, 'properties')
    return get_nodes(container, 'property', filter)


def createProperty(tree, id, **kwargs):
    r"""Creates a single property.

    :param tree: (:py:class:`heimdall.elements.Root`) -- HERA element tree
    :param id: (:py:class:`str`) Property unique identifier
    :param \**kwargs: (:py:class:`dict`, optional) -- Keyword arguments, see below.
    :Keyword arguments:
        * **type** (:py:class:`str`, optional, default: ``text``) -- Property type
        * **name** (:py:class:`str`\ \|\ :py:class:`dict`, optional) -- Human-readable name
        * **description** (:py:class:`str`\ \|\ :py:class:`dict`, optional) -- Human-readable documentation
        * **uri** (:py:class:`list` of :py:class:`str`, optional) -- URI list
    :return: Property that was just created
    :rtype: :py:class:`heimdall.elements.Property`

    This function can be used to add a new property to a database.
    Elements created by ``createProperty`` will always be added to the ``<properties/>`` container element.

    In its simplest usage, ``createProperty`` simply creates a new ``<property id='xxx'/>`` element: ::

      >>> import heimdall
      >>> ...  # create config
      >>> tree = heimdall.getDatabase(config)  # load HERA tree
      >>> heimdall.createProperty(tree, id='xxx')  # create a new property
      >>> # the following child is now added to the properties list:
      >>> # <property id='xxx' />

    Each property identifier must be unique among all properties.
    Thus, calling ``createProperty`` with an ``id`` already in use by another property in the same database will fail.

    Additional supported parameters are ``type``, ``name``, ``description``, and ``uri``.
    Each of these parameters creates appropriate children for the property.
    Here is an example: ::

      >>> import heimdall
      >>> ...  # create config, load HERA tree
      >>> heimdall.createProperty(tree,
      >>>     id='dc:title', name='Title',
      >>>     uri=[
      >>>         'http://purl.org/dc/terms/title',
      >>>         'http://purl.org/dc/terms/alternative',
      >>>         'http://nakala.fr/terms#title'
      >>>     ])
      >>> # the following property is now added to the properties list:
      >>> # <property id='dc:title'>
      >>> #     <type>text</type>
      >>> #     <name>Title</name>
      >>> #     <uri>http://purl.org/dc/terms/title</uri>
      >>> #     <uri>http://purl.org/dc/terms/alternative</uri>
      >>> #     <uri>http://nakala.fr/terms#title</uri>
      >>> # </property>

    Please note that ``name`` and ``description`` can be localized, if they are of type :py:class:`dict` instead of :py:class:`str` (:py:class:`dict` keys are language codes).
    The following example shows property localization: ::

      >>> import heimdall
      >>> ...  # create HERA element tree
      >>> heimdall.createProperty(tree, id='dc:title',
      >>>      name={
      >>>          'de': "Titel",
      >>>          'en': "Title",
      >>>          'fr': "Titre",
      >>>      },
      >>>      description={
      >>>          'en': "A name given to the resource.",
      >>>          'fr': "Un nom donné à la ressource.",
      >>>      })
      >>> # the following XML element is now added to the ``<properties/>`` container element:
      >>> # <property id='dc:title'>
      >>> #     <type>text</type>
      >>> #     <name xml:lang='de'>Titel</name>
      >>> #     <name xml:lang='en'>Title</name>
      >>> #     <name xml:lang='fr'>Titre</name>
      >>> #     <description xml:lang='en'>A name given to the resource.</description>
      >>> #     <description xml:lang='fr'>Un nom donné à la ressource.</description>
      >>> # </property>
    """  # nopep8: E501
    container = create_container(tree, 'properties')
    # Check that property unique id (pid) does not already exist
    if getProperty(tree, lambda n: n.get('id') == id) is not None:
        raise ValueError(f"Property id '{id}' already exists")
    # Create property
    node = Property(id=id)
    node.type = kwargs.get('type', 'text')
    param = kwargs.get('name', None)
    if param is not None:
        node.name = param
    param = kwargs.get('description', None)
    if param is not None:
        node.description = param
    param = kwargs.get('uri', list())
    if type(param) is not list:
        raise TypeError(f"'uri' expected:'list', got:'{type(param).__name__}'")
    node.uri = param
    container.append(node)
    return node


def replaceProperty(property, **kwargs):
    """Replaces an existing property.

    .. ERROR::
       This feature is not implemented yet.
       This can be either due to lack of resources, lack of demand, because it wouldn't be easily maintenable, or any combination of these factors.

       Interested readers can submit a request to further this topic.
       See the ``CONTRIBUTING.rst`` file at the root of the repository for details.
    """  # nopep8: E501
    raise ValueError("TODO: Not Implemented")


def updateProperty(property, **kwargs):
    r"""Updates a single property.

    :param property: (:py:class:`heimdall.elements.Property`) -- HERA property
    :param \**kwargs: (:py:class:`dict`, optional) -- Keyword arguments, see below.
    :Keyword arguments:
        * **id** (:py:class:`str`, optional) -- Property identifier
        * **type** (:py:class:`str`, optional) -- Property type
        * **name** (:py:class:`str`\ \|\ :py:class:`dict`, optional) -- Human-readable name
        * **description** (:py:class:`str`\ \|\ :py:class:`dict`, optional) -- Human-readable documentation
        * **uri** (:py:class:`list` of :py:class:`str`, optional) -- URI list
    :return: Property that was just updated
    :rtype: :py:class:`heimdall.elements.Property`

    This function can be used to update an existing property.

    Any or all of ``updateProperty`` parameters can be used to update ``property``: ::

      >>> import heimdall
      >>> ...  # load HERA tree
      >>> p = heimdall.getProperty(tree, lambda x: x.get('id') == 'title')
      >>> heimdall.updateProperty(p, name='Label')

    Please note that ``name`` and ``description`` can be localized, if they are of type :py:class:`dict` instead of :py:class:`str` (:py:class:`dict` keys are language codes).
    The following example shows property localization: ::

      >>> import heimdall
      >>> ...  # load HERA element tree
      >>> p = heimdall.getProperty(tree, ...)
      >>> heimdall.updateProperty(p,
      >>>      name={
      >>>          'de': "Titel",
      >>>          'en': "Title",
      >>>          'fr': "Titre",
      >>>      },
      >>>      description={
      >>>          'en': "A name given to the resource.",
      >>>          'fr': "Un nom donné à la ressource.",
      >>>      })
    """  # nopep8: E501
    VALUES = ['id', ]
    CHILDREN = ['type', 'name', 'description', 'uri', ]
    _maybe_update_values(property, VALUES, **kwargs)
    _maybe_update_children(property, CHILDREN, **kwargs)
    return property


def deleteProperty(tree, filter):
    """Deletes a single property.

    :param tree: (:py:class:`heimdall.elements.Root`) -- HERA element tree
    :param filter: (:py:class:`function`) -- Filtering function
    :return: None
    :rtype: :py:class:`NoneType`

    This function can be used to delete properties from ``tree``.
    It performs the property deletion "in place"; in other words, parameter ``tree`` is directly modified, and this function returns nothing.

    If ``filter`` returns no result, this function does nothing, and does not raise any error.

    This method doesn't delete any metadata, nor any attribute, referencing the deleted property.
    All elements referencing this property, if any, will become invalid.
    They should either be deleted or fixed accordingly.

    Usage ::

      >>> import heimdall
      >>> ...  # create config, load HERA tree
      >>> # delete a property using its unique id
      >>> heimdall.deleteProperty(tree, lambda e: e.get('id') == '42')
    """  # nopep8: E501
    nodes = getProperties(tree, filter)
    container = get_container(tree, 'properties')
    for node in nodes:
        container.remove(node)


__copyright__ = "Copyright the pyHeimdall contributors."
__license__ = 'AGPL-3.0-or-later'
__all__ = [
    'getProperty', 'getProperties',
    'createProperty', 'deleteProperty',
    'replaceProperty', 'updateProperty',
    '__copyright__', '__license__',
    ]



# -*- coding: utf-8 -*-
import heimdall
import os as _os
import re as _re
from sys import version_info as _py
from urllib.parse import urlparse
from ..decorators import get_database, create_database
LENGTH_PREFIX = 'len:'


@create_database(['sql:mariadb', 'sql:mysql', ])
def dump(tree, url, **options):
    r"""Serialize a HERA elements tree into a MySQL dump like file

    :param tree: HERA elements tree
    :param url: Path of the MySQL dump file to create
    :param \**options: (optional) Keyword arguments, see below.
    :Keyword arguments:
        * **allow_multivalues** (``bool``) -- (optional, default: ``True``)
          If ``True`` multiple values for the same entity attribute will be
          put in the same SQL column, separated by ``multivalue_separator``.
          If ``False``, this function will raise an error
          should any attribute have a ``max`` greater than 1.
        * **multivalue_separator** (``str``) -- (optional, default: ``,``)
          "Character" separating multiple values for the same attribute
          in a given entity; used only if ``allow_multivalues`` is ``True``.
    """
    path = url
    if _os.path.exists(path):
        raise ValueError(f"File '{path}' already exists")

    allow = options.get('allow_multivalues', False)
    allow = True  # TODO
    comma = options.get('multivalue_separator', ',')
    with open(path, 'w', newline='') as f:
        f.write(f"{_pre_everything()}\n")
        for entity in heimdall.getEntities(tree):
            eid = entity.get('id')
            attributes = heimdall.getAttributes(entity)
            items = heimdall.getItems(tree, lambda n: n.get('eid') == eid)
            f.write(f"\n{_pre_create_table(eid)}\n")
            f.write(f"{_create_table(tree, eid, attributes, allow)}\n")
            f.write(f"{_pre_dump_data(eid)}\n")
            f.write(f"{_dump_data(items, eid, attributes, comma)}\n")
            f.write(f"{_post_dump_data()}\n")


def _pre_everything():
    text = "-- SQL dump, from a HERA database, using pyHeimdall\n"
    text += "-- ------------------------------------------------------\n"
    text += f"-- pyHeimdall version	{heimdall.__version__}"
    return text


def _pre_create_table(eid, drop=True):
    comment = f"--\n-- Table structure for table `{eid}`\n--"
    drop_table = f"\nDROP TABLE IF EXISTS `{eid}`;" if drop else ""
    return comment + drop_table


def _create_table(tree, eid, attributes, allow_multivalues):
    text = f"CREATE TABLE `{eid}` (\n"
    for attribute in attributes:
        (aid, pid) = _get_identifiers(attribute)
        if aid is None and pid is None:
            continue  # this unused attribute won't be part of the dump
        _id = aid if aid is not None else pid
        max = attribute.get('max')
        max = int(max) if max else None
        if not allow_multivalues and (max is None or max > 1):
            fault = f"{eid}.{aid}.max={max}"
            raise ValueError(f"Repeatable attributes not supported ({fault})")
        min = int(attribute.get('min') or 0)
        min = "NOT NULL" if min > 0 else "DEFAULT NULL"
        type = _get_type(tree, attribute)
        text += f"`{_id}` {type} {min},\n"
    if text.endswith(',\n'):
        text = text[:-2] + '\n'  # remove last comma
    text += ") ENGINE=InnoDB DEFAULT CHARSET=utf8 COLLATE=utf8_general_ci;"
    return text


def _get_identifiers(attribute):
    return (attribute.get('id'), attribute.get('pid'))


def _get_type(tree, attribute):
    type_node = heimdall.util.get_node(attribute, 'type')
    if type_node is not None:
        type_ = type_node.text
    else:
        pid = attribute.get('pid')
        property_ = heimdall.getProperty(tree, lambda n: n.get('id') == pid)
        type_node = heimdall.util.get_node(attribute, 'type')
        if type_node is not None:
            type_ = type_node.text
        else:
            type_ = 'text'
    if type_ == 'datetime':
        return 'date'
    if type_ == 'number':
        return 'int'  # TODO unsigned
    rules = heimdall.util.get_nodes(attribute, 'rule')
    length = 255
    for rule in rules:
        if rule.text.startswith(LENGTH_PREFIX):
            length = int(rule.text[len(LENGTH_PREFIX):])
    if length > 1:
        return f'varchar({length})'
    return 'char(1)'


def _pre_dump_data(eid, lock=True):
    comment = f"--\n-- Dumping data for table `{eid}`\n--"
    lock_table = f"\nLOCK TABLE `{eid}` WRITE;" if lock else ""
    return comment + lock_table


def _dump_data(items, eid, attributes, multivalues_separator):
    if len(items) == 0:
        return f"-- No data in table `{eid}`"
    text = f"INSERT INTO `{eid}` VALUES "
    dumps = list()
    for item in items:
        values = list()
        for attribute in attributes:
            values.append(_dump_value(item, attribute, multivalues_separator))
        dumps.append(f"({','.join(values)})")
    text += ','.join(dumps)
    text += ";"
    return text


def _dump_value(item, attribute, multivalues_separator):
    (aid, pid) = _get_identifiers(attribute)
    values = list()
    if aid is not None:
        if pid is not None:
            values = heimdall.getValues(item, pid=pid, aid=aid)
        else:
            values = heimdall.getValues(item, aid=aid)
    else:
        if pid is not None:
            values = heimdall.getValues(item, pid=pid)
    values = [f"'{v}'" for v in values]
    if len(values) < 1:
        return 'NULL'
    return multivalues_separator.join(values)


def _post_dump_data(lock=True):
    return "UNLOCK TABLES;"



# -*- coding: utf-8 -*-
from .heimdall import CONNECTORS as _CONNECTORS
from functools import wraps


def get_database(format):
    # just two lines to support syntax: @get_database('format')
    # as long as syntax: @get_database(['format1', 'format2'])
    if type(format) is not list:
        format = [format, ]

    def decorator(function):
        # register ``function`` as an IN connector
        for f in format:
            _CONNECTORS['get_database'][f] = function

        @wraps(function)
        def wrapper(*args, **kwargs):
            # call ``function`` and return its result
            return function(*args, **kwargs)
        return wrapper
    return decorator


def create_database(format):
    # just two lines to support syntax: @create_database('format')
    # as long as syntax: @create_database(['format1', 'format2'])
    if type(format) is not list:
        format = [format, ]

    def decorator(function):
        # register ``function`` as an OUT connector
        for f in format:
            _CONNECTORS['create_database'][f] = function

        @wraps(function)
        def wrapper(*args, **kwargs):
            # call ``function`` and return its result
            return function(*args, **kwargs)
        return wrapper
    return decorator



# -*- coding: utf-8 -*-

"""
Provides utility functions around HERA attributes.

:copyright: The pyHeimdall contributors.
:licence: Afero GPL, see LICENSE for more details.
:SPDX-License-Identifier: AGPL-3.0-or-later
"""
import heimdall
from heimdall.util.tree import get_node, set_language
from xml.etree.ElementTree import Element


def refactor_relationship(tree, relationship, eid, euid, pid, cleanup=True):
    """This function merges an entity into two others.

    :param tree: (:py:class:`heimdall.elements.Root`) -- HERA element tree
    :param relationship: (:py:class:`dict`) -- Description of the relationship entity we want to merge
    :param eid: (:py:class:`str`) -- Identifier of the HERA entity in ``tree`` to which we'll add ``relationship['target']`` as a repeatable field
    :param euid: (:py:class:`str`) -- Property identifier of an attribute which can uniquely identify entity ``eid``
    :param pid: (:py:class:`str`) -- Property identifier of the new attribute to create in entity ``eid`` for ``relationship['target']``
    :param cleanup: (:py:class:`bool`, optional, default: ``True``) -- If ``True`` (default), items in ``tree`` representing ``relationship['eid']`` will be deleted
    :return: None
    :rtype: :py:class:`NoneType`

    This function merges an entity that is in fact a relationship between two other entities, so the corresponding items become repeatable metadata of one of these two entities.
    It performs its modifications "in place".
    In other words, parameter ``tree`` is directly modified,
    and this function returns nothing.

    **Usage example**

    Let's say ``tree`` initialy contains three entities:

    * '`person`' to represents persons.
      Items of entity '`person`' can be uniquely identified by their '`name`' metadata; attribute '`name`' is the concrete use by '`person`' of property '`title`';
    * '`pet`' to represent pets;
    * '`owners2pets`' representing who owns who, via its two only attributes '`owner`' and '`pet`', as a person can have multiple pets.

    This kind of "indirect relation" entity is often found in databases whose original implementations handle repeatable fields like this.
    Problem is: on many databases, there are many of this kind of entity, they clutter the database schema only because of the limitations of the original implementation.
    This makes the whole database intention less readable, thus less accessible, thus more difficult to understand and reuse.

    Here is how a single call to ``refactor_relationship`` can make things simpler: ::

      >>> import heimdall
      >>> tree = heimdall.getDatabase(...)  # load HERA tree
      >>> refactor_relationship(tree,
      >>>     relationship={
      >>>         'eid': 'owners2pets',  # entity we want to get rid of
      >>>         'source': 'owner',  # pointer to 'person'
      >>>         'target': 'pet',  # metadata we want to keep
      >>>         },
      >>>     eid='person',  # entity the new attribute will be added to
      >>>     euid='title',  # unique identifier property for entity 'person'
      >>>     pid='pet')  # property id for the new attribute

    | After ``refactor_relationship`` is called, ``tree`` only contain two entities ('`person`' and '`pet`'), and entity '`person`' has gained a new repeatable field, '`pet`'.
    | As ``cleanup`` was left to ``True``, items previously belonging to now deleted entity '`owners2pets`' are deleted too.
      No data is lost however, as these items only interesting metadata is now hold by metadata '`pet`' in items of entity '`person`'.

    :see also: | :py:class:`heimdall.util.merge_properties` can be used to merge together properties that are semantically the same.
               | :py:class:`heimdall.util.merge_l10n_attributes` can be used to merge together attributes that are translations of one another.
               | :py:class:`heimdall.util.delete_unused_properties` can be used to cleanup what is left after doing this kind of "tree improvement" operations.
    """  # nopep8: E501
    R_ID = relationship['eid']
    R_SOURCE = relationship['source']
    R_TARGET = relationship['target']
    # TODO only create property if not exists
    p = heimdall.createProperty(tree, id=pid)
    e = heimdall.getEntity(tree, lambda n: n.get('id') == eid)
    # TODO only create attribute if not exists
    a = heimdall.createAttribute(e, pid=pid)
    # iterate over all items belonging to the relationship entity
    items = heimdall.getItems(tree, lambda n: n.get('eid') == R_ID)
    for old in items:
        source = heimdall.getMetadata(old, pid=R_SOURCE)[0].text
        target = heimdall.getMetadata(old, pid=R_TARGET)[0].text

        def is_source(item):
            is_of_entity = item.get('eid') == eid
            has_unique_id = False
            # NOTE: Unique id shouldn't be a repeatable attribute,
            # but we know what real world data looks like.
            # Thus, try to not break in this edge case, and
            # let's hope our caller knows what she does.
            for v in heimdall.getValues(item, pid=euid):
                has_unique_id = has_unique_id or (v == source)
            return is_of_entity and has_unique_id

        # get the item which must contain the new repeatable metadata
        now = heimdall.getItem(tree, is_source)
        metadata = Element('metadata', pid=pid)
        metadata.text = target
        now.append(metadata)
        if cleanup:
            # delete `old` relationship item, because it is
            # now represented by new metadata in item `now`
            heimdall.deleteItem(tree, lambda n: n == old)
    if cleanup:
        # delete the `relationship` entity, as there are no more items using it
        heimdall.deleteEntity(tree, lambda n: n.get('id') == R_ID)


def merge_l10n_attributes(
        tree, eid, languages,
        pid, aid,
        cleanup=True, update_items=True
        ):
    """Merges attributes that are in fact translations of one another.

    :param tree: (:py:class:`heimdall.elements.Root`) -- HERA element tree
    :param eid: (:py:class:`str`) -- Identifier of the HERA entity in ``tree`` containing all attributes to merge
    :param languages: (:py:class:`dict`) -- :py:class:`dict` containing ``eid``'s "translated" attributes identifiers as keys and their actual language codes as values
    :param pid: (:py:class:`str`) -- Property identifier of "merged" attribute
    :param aid: (:py:class:`str`) -- Attribute identifier of "merged" attribute; if ``aid`` does not identify an attribute already present in ``entity``, a new attribute of this identifier and of type ``text`` will be created
    :param cleanup: (:py:class:`bool`, optional, default: ``True``) -- If ``True`` (default), attributes in ``languages`` will be removed from entity identified ``eid`` after they are merged; attribute identified by ``aid`` is of course never removed
    :param update_items: (:py:class:`bool`, optional, default: ``True``) -- If ``True``, metadata items in ``tree`` using attributes listed in ``languages`` will be updated with the new ``pid`` and ``aid``.
    :return: None
    :rtype: :py:class:`NoneType`

    This function merges attributes that are in fact translations of one another in the same entity into one (possibly new) "merged" attribute.
    It performs its modifications "in place".
    In other words, parameter ``tree`` is directly modified,
    and this function returns nothing.

    **Usage example**

    Let's say ``tree`` initialy contains the following '`person`' entity to represents persons: ::

      >>> <entity id='person'>
      >>>   <attribute pid='name_de' id='person.name_de_attr'>
      >>>     <name>Personenname</name>
      >>>   </attribute>
      >>>   <attribute pid='name_en' id='person.name_en_attr'>
      >>>     <name>Name</name>
      >>>     <description>Name of the person</description>
      >>>   </attribute>
      >>>   <attribute pid='name_fr' id='person.name_fr_attr'>
      >>>     <name>Nom</name>
      >>>     <description>Nom de la personne</description>
      >>>   </attribute>
      >>> </entity>

    Such duplication of attributes is often observed in databases implementations where internationalization and repeatability of values are not features that are available or generally used.
    It makes the resulting entity unwieldy, as semantic relations between metadata values are lost in their syntax.
    The resulting database can be seen as less interoperable, as it requires ad-hoc shenahigans to restore proper internationalization.

    The function ``merge_l10n_attributes`` provides an alternative.
    To merge those three attributes of entity '`person`', one could do the following: ::

      >>> import heimdall
      >>> tree = heimdall.getDatabase(...)  # load HERA tree
      >>> # (1) First, give ourselves a new property
      >>> #     for merging human-readable names:
      >>> heimdall.createProperty(
      >>>     tree, 'dc:title', type='text',
      >>>     name={'en': "Title", 'fr': "Titre", },
      >>>     description={
      >>>         'en': "A name given to the resource.",
      >>>         'fr': "Nom de la ressource.",
      >>>         },
      >>>     uri=[
      >>>         'http://purl.org/dc/terms/title',
      >>>         'http://datacite.org/schema/kernel-4/title',
      >>>         ],
      >>>     )
      >>> # (2) Retrieve the 'person' entity we want to work on:
      >>> e = heimdall.getEntity(tree, lambda n: n.get('id') == 'person')
      >>> # (3) Merge all the things !
      >>> merge_l10n_attributes(tree, e, {
      >>>     'person.name_de_attr': 'de',
      >>>     'person.name_en_attr': 'en',
      >>>     'person.name_fr_attr': 'fr',
      >>>     },
      >>>     aid='person.name',
      >>>     pid='dc:title')

    After running the above code, entity '`person`' looks as follows:

      >>> <entity id='person'>
      >>>   <attribute pid='dc:title' id='person.name'>
      >>>     <name xml:lang='de'>Personenname</name>
      >>>     <name xml:lang='en'>Name</name>
      >>>     <name xml:lang='fr'>Nom</name>
      >>>     <description xml:lang='en'>Name of the person</description>
      >>>     <description xml:lang='fr'>Nom de la personne</description>
      >>>   </attribute>
      >>> </entity>

    One can argue entity '`person`' is now leaner and easier to work with, as long as more accessible and reusable, as it is based on scientific standards like `Dublin Core™ <https://www.dublincore.org/specifications/dublin-core/>`_, and doesn't separate anymore attributes that were "the same in a different language".

    **Impact on items**

    | This is further enforced by the update of items in ``tree`` allowed by setting ``update_items`` parameter to ``True`` (the default).
    | In the example above, before calling ``merge_l10n_attributes``, items would look like that: ::

      >>> <item eid='person'>
      >>>   <metadata pid='name_de' aid='person.name_de_attr'>Wilhem</metadata>
      >>>   <metadata pid='name_en' aid='person.name_en_attr'>William</metadata>
      >>>   <metadata pid='name_fr' aid='person.name_fr_attr'>Guillaume</metadata>
      >>> </item>

    | Syntactically speaking, those are three entirely different metadata.
    | After calling ``merge_l10n_attributes``, items would rather contain `one single` metadata, `localized in three languages`, and would look like this: ::

      >>> <item eid='person'>
      >>>   <metadata pid='title' aid='person.name' xml:lang='de'>Wilhem</metadata>
      >>>   <metadata pid='title' aid='person.name' xml:lang='en'>William</metadata>
      >>>   <metadata pid='title' aid='person.name' xml:lang='fr'>Guillaume</metadata>
      >>> </item>

    :see also: | :py:class:`heimdall.util.merge_properties` can be used to merge together properties that are semantically the same (from a scientific metadata standard point of view), although named differently in different entities of the imported ``tree``.
               | :py:class:`heimdall.util.refactor_relationship` can be used to merge a whole entity into two others.
               | :py:class:`heimdall.util.delete_unused_properties` can be used to cleanup what is left after doing this kind of "tree improvement" operations.
               | :py:class:`heimdall.util.get_language` and :py:class:`heimdall.util.set_language` are useful functions related to HERA elements localization.
    """  # nopep8: E501
    entity = heimdall.getEntity(tree, lambda n: n.get('id') == eid)
    if entity is None:
        raise ValueError(f"Entity '{eid}' doesn't exist")
    (base, new) = _get_or_create_attribute(entity, aid, pid)
    base_aid = base.get('id')
    base_pid = base.get('pid')
    _update_attribute(base, entity, languages)
    if new:
        entity.append(base)
    # * ... if `cleanup` delete `language` attributes (except base_atr ofc)
    # * ... if `update_items`, update_items to use id = base_a.id / base_aid
    items = []
    if update_items:
        items = heimdall.getItems(tree, lambda n: n.get('eid') == eid)
    for aid, language in languages.items():
        if cleanup and (aid != base_aid):
            attr = heimdall.getAttribute(entity, lambda a: a.get('id') == aid)
            entity.remove(attr)
        for item in items:
            for metadata in heimdall.getMetadata(item, aid=aid):
                metadata.set('aid', base_aid)
                metadata.set('pid', base_pid)
                set_language(metadata, language)
    return base


def _get_or_create_attribute(entity, aid, pid):
    """Retrieve attribute from ``entity``:
    - if one ``aid`` is in ``entity``, return it
    - if ONLY one ``pid`` is in ``entity``, return it
    - else return a new Attribute element, which is NOT added to ``entity`` yet
    """
    new = False
    base = heimdall.getAttribute(entity, lambda n: n.get('id') == aid)
    if base is None:
        base = heimdall.elements.Attribute(id=aid)
        new = True
    if pid is not None:
        base.set('pid', pid)
    return (base, new)


def _update_attribute(base, entity, languages):
    """Update ``base`` attribute with type and pid of
    attributes listed in ``langages``, while checking
    these type and pid are the same.
    """
    base_aid = base.get('id')
    types = list()
    for aid, language in languages.items():
        a = heimdall.getAttribute(entity, lambda n: n.get('id') == aid)
        if a is None:
            raise ValueError(f"Unknown attribute identifier '{aid}'")
        _update_language(base, a, 'name', language)
        _update_language(base, a, 'description', language)
        if a.type is not None:
            types.append(a.type)
    types = list(set(types))
    if len(types) > 1:
        raise ValueError("Attributes don't all have the same type")
    if len(types) > 0:
        base.type = types[0]
    # TODO: check compatibility of types[0] (if exists)
    #       with underlying property
    # TODO: set min and max
    return base


def _update_language(base, attribute, key, language):
    values = getattr(base, key, dict())
    old_values = getattr(attribute, key, dict())
    for k, v in old_values.items():
        if k is not None and k != language:
            continue
        value = values.get(language, list())
        value = list(set(value + v))
        values[language] = value
    values.pop(None, None)
    setattr(base, key, values)
    return values


__copyright__ = "Copyright the pyHeimdall contributors."
__license__ = 'AGPL-3.0-or-later'



# -*- coding: utf-8 -*-

from heimdall.util.tree import (
    get_node, get_nodes, get_root,
    create_node, create_nodes,
    get_language, set_language,
    )
from heimdall.util.entities import (
    update_entities,
    )
from heimdall.util.attributes import (
    merge_l10n_attributes,
    refactor_relationship,
    )
from heimdall.util.properties import (
    delete_unused_properties,
    merge_properties,
    )


__all__ = [
    'update_entities',
    'merge_l10n_attributes',
    'refactor_relationship',
    'delete_unused_properties', 'merge_properties',

    '__copyright__', '__license__',
    ]
__copyright__ = "Copyright the pyHeimdall contributors."
__license__ = 'AGPL-3.0-or-later'
